
import torch
import torch.nn as nn
from torchvision.utils import save_image
import numpy as np
import os
import timeit
from PIL import Image




#-------------------------------------------------------------
def train_net_embed(net, net_name, trainloader, testloader, epochs=200, resume_epoch = 0, lr_base=0.01, lr_decay_factor=0.1, lr_decay_epochs=[80, 140], weight_decay=1e-4, path_to_ckpt = None):

    ''' learning rate decay '''
    def adjust_learning_rate_1(optimizer, epoch):
        """decrease the learning rate """
        lr = lr_base

        num_decays = len(lr_decay_epochs)
        for decay_i in range(num_decays):
            if epoch >= lr_decay_epochs[decay_i]:
                lr = lr * lr_decay_factor
            #end if epoch
        #end for decay_i
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr

    net = net.cuda()
    criterion = nn.MSELoss()
    optimizer = torch.optim.SGD(net.parameters(), lr = lr_base, momentum= 0.9, weight_decay=weight_decay)

    # resume training; load checkpoint
    if path_to_ckpt is not None and resume_epoch>0:
        save_file = path_to_ckpt + "/embed_x2y_ckpt_in_train/embed_x2y_checkpoint_epoch_{}.pth".format(resume_epoch)
        checkpoint = torch.load(save_file)
        net.load_state_dict(checkpoint['net_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        torch.set_rng_state(checkpoint['rng_state'])
    #end if

    start_tmp = timeit.default_timer()
    for epoch in range(resume_epoch, epochs):
        net.train()
        train_loss = 0
        adjust_learning_rate_1(optimizer, epoch)
        for _, (batch_train_images, batch_train_labels) in enumerate(trainloader):

            # batch_train_images = nn.functional.interpolate(batch_train_images, size = (299,299), scale_factor=None, mode='bilinear', align_corners=False)

            batch_train_images = batch_train_images.type(torch.float).cuda()
            batch_train_labels = batch_train_labels.type(torch.float).view(-1,1).cuda()

            #Forward pass
            outputs, _ = net(batch_train_images)
            loss = criterion(outputs, batch_train_labels)

            #backward pass
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            train_loss += loss.cpu().item()
        #end for batch_idx
        train_loss = train_loss / len(trainloader)

        if testloader is None:
            print('Train net_x2y for embedding: [epoch %d/%d] train_loss:%f Time:%.4f' % (epoch+1, epochs, train_loss, timeit.default_timer()-start_tmp))
        else:
            net.eval()  # eval mode (batchnorm uses moving mean/variance instead of mini-batch mean/variance)
            with torch.no_grad():
                test_loss = 0
                for batch_test_images, batch_test_labels in testloader:
                    batch_test_images = batch_test_images.type(torch.float).cuda()
                    batch_test_labels = batch_test_labels.type(torch.float).view(-1,1).cuda()
                    outputs,_ = net(batch_test_images)
                    loss = criterion(outputs, batch_test_labels)
                    test_loss += loss.cpu().item()
                test_loss = test_loss/len(testloader)

                print('Train net_x2y for label embedding: [epoch %d/%d] train_loss:%f test_loss:%f Time:%.4f' % (epoch+1, epochs, train_loss, test_loss, timeit.default_timer()-start_tmp))

        #save checkpoint
        if path_to_ckpt is not None and (((epoch+1) % 50 == 0) or (epoch+1==epochs)):
            save_file = path_to_ckpt + "/embed_x2y_ckpt_in_train/embed_x2y_checkpoint_epoch_{}.pth".format(epoch+1)
            os.makedirs(os.path.dirname(save_file), exist_ok=True)
            torch.save({
                    'epoch': epoch,
                    'net_state_dict': net.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                    'rng_state': torch.get_rng_state()
            }, save_file)
    #end for epoch

    return net




###################################################################################
class label_dataset(torch.utils.data.Dataset):
    def __init__(self, labels):
        super(label_dataset, self).__init__()

        self.labels = labels
        self.n_samples = len(self.labels)

    def __getitem__(self, index):

        y = self.labels[index]
        return y

    def __len__(self):
        return self.n_samples


def train_net_y2h(unique_labels_norm, net_y2h, net_embed, epochs=500, lr_base=0.01, lr_decay_factor=0.1, lr_decay_epochs=[150, 250, 350], weight_decay=1e-4, batch_size=128):
    '''
    unique_labels_norm: an array of normalized unique labels
    '''

    ''' learning rate decay '''
    def adjust_learning_rate_2(optimizer, epoch):
        """decrease the learning rate """
        lr = lr_base

        num_decays = len(lr_decay_epochs)
        for decay_i in range(num_decays):
            if epoch >= lr_decay_epochs[decay_i]:
                lr = lr * lr_decay_factor
            #end if epoch
        #end for decay_i
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr


    assert np.max(unique_labels_norm)<=1 and np.min(unique_labels_norm)>=0
    trainset = label_dataset(unique_labels_norm)
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=True)

    net_embed.eval()
    net_h2y=net_embed.module.h2y #convert embedding labels to original labels
    optimizer_y2h = torch.optim.SGD(net_y2h.parameters(), lr = lr_base, momentum= 0.9, weight_decay=weight_decay)

    start_tmp = timeit.default_timer()
    for epoch in range(epochs):
        net_y2h.train()
        train_loss = 0
        adjust_learning_rate_2(optimizer_y2h, epoch)
        for _, batch_labels in enumerate(trainloader):

            batch_labels = batch_labels.type(torch.float).view(-1,1).cuda()

            # generate noises which will be added to labels
            batch_size_curr = len(batch_labels)
            batch_gamma = np.random.normal(0, 0.2, batch_size_curr)
            batch_gamma = torch.from_numpy(batch_gamma).view(-1,1).type(torch.float).cuda()

            # add noise to labels
            batch_labels_noise = torch.clamp(batch_labels+batch_gamma, 0.0, 1.0)

            #Forward pass
            batch_hiddens_noise = net_y2h(batch_labels_noise)
            batch_rec_labels_noise = net_h2y(batch_hiddens_noise)

            loss = nn.MSELoss()(batch_rec_labels_noise, batch_labels_noise)

            #backward pass
            optimizer_y2h.zero_grad()
            loss.backward()
            optimizer_y2h.step()

            train_loss += loss.cpu().item()
        #end for batch_idx
        train_loss = train_loss / len(trainloader)

        print('\n Train net_y2h: [epoch %d/%d] train_loss:%f Time:%.4f' % (epoch+1, epochs, train_loss, timeit.default_timer()-start_tmp))
    #end for epoch

    return net_y2h



import cv2
import math
import numpy as np
from PIL import Image
from skimage.metrics import structural_similarity as ssim


# MSE
def _MSE(image_1, image_2):
    array_1 = np.array(image_1) / 255
    array_2 = np.array(image_2) / 255
    n, m, _ = array_1.shape
    return np.sum((array_1 - array_2) ** 2) / (n * m * 3)


def MSE(images):
    MSE_sum = 0
    n = len(images)
    for i in range(n):
        for j in range(i + 1, n):
            MSE_sum += _MSE(images[i], images[j])
    return MSE_sum / (n * (n - 1) / 2)


# SSIM


def _SSIM(image_1, image_2):

    array_1 = np.array(image_1)
    array_2 = np.array(image_2)
    score, _ = ssim(array_1, array_2, channel_axis=-1, full=True)
    return score


def SSIM(images):
    SSIM_sum = 0
    n = len(images)
    for i in range(n):
        for j in range(i + 1, n):
            SSIM_sum += _SSIM(images[i], images[j])
    return SSIM_sum / (n * (n - 1) / 2)


# PSNR
def _PSNR(image_1, image_2):
    mse = _MSE(image_1, image_2)
    if mse == 0:
        return float("inf")
    else:
        return 20 * math.log10(1.0 / math.sqrt(mse))


def PSNR(images):
    PSNR_sum = 0
    n = len(images)
    for i in range(n):
        for j in range(i + 1, n):
            PSNR_sum += _PSNR(images[i], images[j])
    return PSNR_sum / (n * (n - 1) / 2)


# Similarity
_method_dict = {"MSE": MSE, "SSIM": SSIM, "PSNR": PSNR}


def similarity(images, method):
    return _method_dict[method](images)


__all__ = ["similarity"]



'''
codes are based on
@article{
zhang2018mixup,
title={mixup: Beyond Empirical Risk Minimization},
author={Hongyi Zhang, Moustapha Cisse, Yann N. Dauphin, David Lopez-Paz},
journal={International Conference on Learning Representations},
year={2018},
url={https://openreview.net/forum?id=r1Ddp1-Rb},
}
'''


import torch
import torch.nn as nn
import torch.nn.functional as F

NC = 3
IMG_SIZE = 128


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet_regre_eval(nn.Module):
    def __init__(self, block, num_blocks, nc=NC, ngpu = 1, feature_layer='f3'):
        super(ResNet_regre_eval, self).__init__()
        self.in_planes = 64
        self.ngpu = ngpu
        self.feature_layer=feature_layer

        self.block1 = nn.Sequential(
            nn.Conv2d(nc, 64, kernel_size=3, stride=1, padding=1, bias=False),  # h=h
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2,2), #h=h/2 64
            self._make_layer(block, 64, num_blocks[0], stride=2),  # h=h/2 32
        )
        self.block2 = self._make_layer(block, 128, num_blocks[1], stride=2) # h=h/2 16
        self.block3 = self._make_layer(block, 256, num_blocks[2], stride=2) # h=h/2 8
        self.block4 = self._make_layer(block, 512, num_blocks[3], stride=2) # h=h/2 4

        self.pool1 = nn.AvgPool2d(kernel_size=4)
        if self.feature_layer == 'f2':
            self.pool2 = nn.AdaptiveAvgPool2d((2,2))
        elif self.feature_layer == 'f3':
            self.pool2 = nn.AdaptiveAvgPool2d((2,2))
        else:
            self.pool2 = nn.AdaptiveAvgPool2d((1,1))

        linear_layers = [
                nn.Linear(512*block.expansion, 128),
                nn.BatchNorm1d(128),
                nn.ReLU(),
                nn.Linear(128, 128),
                nn.BatchNorm1d(128),
                nn.ReLU(),
                nn.Linear(128, 1),
                # nn.Sigmoid()
                nn.ReLU(),
            ]
        self.linear = nn.Sequential(*linear_layers)


    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):

        if x.is_cuda and self.ngpu > 1:
            ft1 = nn.parallel.data_parallel(self.block1, x, range(self.ngpu))
            ft2 = nn.parallel.data_parallel(self.block2, ft1, range(self.ngpu))
            ft3 = nn.parallel.data_parallel(self.block3, ft2, range(self.ngpu))
            ft4 = nn.parallel.data_parallel(self.block4, ft3, range(self.ngpu))
            out = nn.parallel.data_parallel(self.pool1, ft4, range(self.ngpu))
            out = out.view(out.size(0), -1)
            out = nn.parallel.data_parallel(self.linear, out, range(self.ngpu))
        else:
            ft1 = self.block1(x)
            ft2 = self.block2(ft1)
            ft3 = self.block3(ft2)
            ft4 = self.block4(ft3)
            out = self.pool1(ft4)
            out = out.view(out.size(0), -1)
            out = self.linear(out)

        if self.feature_layer == 'f2':
            ext_features = self.pool2(ft2)
        elif self.feature_layer == 'f3':
            ext_features = self.pool2(ft3)
        else:
            ext_features = self.pool2(ft4)

        ext_features = ext_features.view(ext_features.size(0), -1)

        return out, ext_features


def ResNet18_regre_eval(ngpu = 1):
    return ResNet_regre_eval(BasicBlock, [2,2,2,2], ngpu = ngpu)

def ResNet34_regre_eval(ngpu = 1):
    return ResNet_regre_eval(BasicBlock, [3,4,6,3], ngpu = ngpu)

def ResNet50_regre_eval(ngpu = 1):
    return ResNet_regre_eval(Bottleneck, [3,4,6,3], ngpu = ngpu)

def ResNet101_regre_eval(ngpu = 1):
    return ResNet_regre_eval(Bottleneck, [3,4,23,3], ngpu = ngpu)

def ResNet152_regre_eval(ngpu = 1):
    return ResNet_regre_eval(Bottleneck, [3,8,36,3], ngpu = ngpu)


if __name__ == "__main__":
    net = ResNet34_regre_eval(ngpu = 1).cuda()
    x = torch.randn(4,NC,IMG_SIZE,IMG_SIZE).cuda()
    out, features = net(x)
    print(out.size())
    print(features.size())



import torch
from torch import nn



class encoder(nn.Module):
    def __init__(self, dim_bottleneck=512, ch=64):
        super(encoder, self).__init__()
        self.ch = ch
        self.dim_bottleneck = dim_bottleneck

        self.conv = nn.Sequential(
            nn.Conv2d(3, ch, kernel_size=4, stride=2, padding=1), #h=h/2; 128
            nn.BatchNorm2d(ch),
            nn.ReLU(),
            nn.MaxPool2d(2,2), #h=h/2; 64
            nn.Conv2d(ch, ch, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch),
            nn.ReLU(),

            nn.Conv2d(ch, ch, kernel_size=4, stride=2, padding=1), #h=h/2; 32
            nn.BatchNorm2d(ch),
            nn.ReLU(),
            nn.Conv2d(ch, ch*2, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*2),
            nn.ReLU(),

            nn.Conv2d(ch*2, ch*2, kernel_size=4, stride=2, padding=1), #h=h/2; 16
            nn.BatchNorm2d(ch*2),
            nn.ReLU(),
            nn.Conv2d(ch*2, ch*4, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*4),
            nn.ReLU(),

            nn.Conv2d(ch*4, ch*4, kernel_size=4, stride=2, padding=1), #h=h/2; 8
            nn.BatchNorm2d(ch*4),
            nn.ReLU(),
            nn.Conv2d(ch*4, ch*8, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*8),
            nn.ReLU(),

            nn.Conv2d(ch*8, ch*8, kernel_size=4, stride=2, padding=1), #h=h/2; 4
            nn.BatchNorm2d(ch*8),
            nn.ReLU(),
            nn.Conv2d(ch*8, ch*16, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*16),
            nn.ReLU(),
        )

        self.linear = nn.Sequential(
            nn.Linear(ch*16*4*4, dim_bottleneck),
            # nn.ReLU()
        )

    def forward(self, x):
        feature = self.conv(x)
        feature = feature.view(-1, self.ch*16*4*4)
        feature = self.linear(feature)
        return feature



class decoder(nn.Module):
    def __init__(self, dim_bottleneck=512, ch=64):
        super(decoder, self).__init__()
        self.ch = ch
        self.dim_bottleneck = dim_bottleneck

        self.linear = nn.Sequential(
            nn.Linear(dim_bottleneck, ch*16*4*4)
        )

        self.deconv = nn.Sequential(
            nn.ConvTranspose2d(ch*16, ch*16, kernel_size=4, stride=2, padding=1), #h=2h; 8
            nn.BatchNorm2d(ch*16),
            nn.ReLU(True),
            nn.Conv2d(ch*16, ch*8, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*8),
            nn.ReLU(),

            nn.ConvTranspose2d(ch*8, ch*8, kernel_size=4, stride=2, padding=1), #h=2h; 16
            nn.BatchNorm2d(ch*8),
            nn.ReLU(True),
            nn.Conv2d(ch*8, ch*4, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*4),
            nn.ReLU(),

            nn.ConvTranspose2d(ch*4, ch*4, kernel_size=4, stride=2, padding=1), #h=2h; 32
            nn.BatchNorm2d(ch*4),
            nn.ReLU(True),
            nn.Conv2d(ch*4, ch*2, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch*2),
            nn.ReLU(),

            nn.ConvTranspose2d(ch*2, ch*2, kernel_size=4, stride=2, padding=1), #h=2h; 64
            nn.BatchNorm2d(ch*2),
            nn.ReLU(True),
            nn.Conv2d(ch*2, ch, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch),
            nn.ReLU(),

            nn.ConvTranspose2d(ch, ch, kernel_size=4, stride=2, padding=1), #h=2h; 128
            nn.BatchNorm2d(ch),
            nn.ReLU(True),
            nn.Conv2d(ch, ch, kernel_size=3, stride=1, padding=1), #h=h
            nn.BatchNorm2d(ch),
            nn.ReLU(),

            nn.ConvTranspose2d(ch, ch, kernel_size=4, stride=2, padding=1), #h=2h; 256
            nn.BatchNorm2d(ch),
            nn.ReLU(True),
            nn.Conv2d(ch, 3, kernel_size=3, stride=1, padding=1), #h=h
            nn.Tanh()
        )

    def forward(self, feature):
        out = self.linear(feature)
        out = out.view(-1, self.ch*16, 4, 4)
        out = self.deconv(out)
        return out


if __name__=="__main__":
    #test

    net_encoder = encoder(dim_bottleneck=512, ch=64).cuda()
    net_decoder = decoder(dim_bottleneck=512, ch=64).cuda()

    x = torch.randn(10, 3, 256, 256).cuda()
    f = net_encoder(x)
    xh = net_decoder(f)
    print(f.size())
    print(xh.size())



import base64
from flask import Flask, send_file
import gdown
import getpass
import h5py
import h5py
import humanize
from importlib import resources
from io import BytesIO
from jinja2 import Template
import numpy as np
import os
from PIL import Image
import subprocess
import time


# view_dataset
def _print_hdf5(name, obj):
    indent = "  " * name.count("/")
    if isinstance(obj, h5py.Dataset):
        print(f"{indent}[Dataset] {name} shape={obj.shape} dtype={obj.dtype}")
    elif isinstance(obj, h5py.Group):
        print(f"{indent}[Group]   {name}")


def view_dataset(dataset_path):
    with h5py.File(dataset_path, "r") as f:
        f.visititems(_print_hdf5)


# visualize_dataset
def _nparray_to_Base64(nparray):
    nparray = np.transpose(nparray, (1, 2, 0))
    image = Image.fromarray(nparray)
    buffered = BytesIO()
    image.save(buffered, format="PNG")
    image_base64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
    return f"data:image/png;base64,{image_base64}"


def visualize_dataset(dataset_path, indexes):
    keys = []
    _data = {}
    with h5py.File(dataset_path, "r") as f:
        for key in f.keys():
            keys.append(key)
            _data[key] = f[key][:]
            if _data[key].dtype.kind == "O":
                _data[key] = _data[key].astype("str")
    keys.append("original_indexes")
    N = len(_data["images"])
    data = []
    for i in indexes:
        item = {}
        for key in keys:
            if key == "images":
                item[key] = _nparray_to_Base64(_data[key][i])
            elif key in ["index_train", "index_valid"]:
                continue
            elif key == "original_indexes":
                item[key] = i
            else:
                item[key] = _data[key][i]
        data.append(item)
    hyperinfo = {
        "template_path": resources.files("myosotis_researches").joinpath(
            "CcGAN", "utils", "src", "template.html"
        ),
        "style_path": resources.files("myosotis_researches").joinpath(
            "CcGAN", "utils", "src", "style.css"
        ),
        "server_path": os.path.abspath(__file__),
        "host": "0.0.0.0",
        "port": "8000",
        "debug": True,
    }
    info = {
        "file": {
            "size": humanize.naturalsize(os.path.getsize(dataset_path), binary=True),
            "created_date": time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(os.path.getctime(dataset_path))
            ),
            "last_edited_date": time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(os.path.getmtime(dataset_path))
            ),
            "path": dataset_path,
            "name": os.path.basename(dataset_path),
        },
        "rows": {"num": N},
        "columns": {"num": len(keys), "list": keys},
        "indexes": indexes,
    }
    all_data = data

    # Local server
    app = Flask(__name__)

    @app.route("/")
    def index():
        with open(hyperinfo["template_path"], "r") as f:
            template = Template(f.read())
        return template.render(hyperinfo=hyperinfo, info=info, datas=all_data)

    @app.route("/style.css")
    def style():
        return send_file(hyperinfo["style_path"])

    app.run(host=hyperinfo["host"], port=hyperinfo["port"], debug=hyperinfo["debug"])


# download_datasets
_file_id_dict = {
    "Ra": "1YVLFAaQ1ldre0ASLSRdV_kPIcAMkwGqO",
    "MNIST": "18249ryrDPsznWWtOkJuALZWCVoZqfibC",
}


def download_datasets(datasets_name, datasets_dir):
    if not os.path.exists(datasets_dir):
        os.makedirs(datasets_dir, exist_ok=True)
    zipped_datasets_path = os.path.join(datasets_dir, f"{datasets_name}.rar")
    file_id = _file_id_dict[datasets_name]
    url = f"https://drive.google.com/uc?id={file_id}"
    gdown.download(url, zipped_datasets_path, quiet=False, use_cookies=False)
    unzip_password = getpass.getpass("Please input the zipped file's password:")
    try:
        cmd = ["unrar", "x", f"-p{unzip_password}", zipped_datasets_path, datasets_dir]
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        print(e)
    os.remove(zipped_datasets_path)


__all__ = ["view_dataset", "visualize_dataset", "download_datasets"]



'''
ResNet-based model to map an image from pixel space to a features space.
Need to be pretrained on the dataset.

if isometric_map = True, there is an extra step (elf.classifier_1 = nn.Linear(512, 32*32*3)) to increase the dimension of the feature map from 512 to 32*32*3. This selection is for desity-ratio estimation in feature space.

codes are based on
@article{
zhang2018mixup,
title={mixup: Beyond Empirical Risk Minimization},
author={Hongyi Zhang, Moustapha Cisse, Yann N. Dauphin, David Lopez-Paz},
journal={International Conference on Learning Representations},
year={2018},
url={https://openreview.net/forum?id=r1Ddp1-Rb},
}
'''


import torch
import torch.nn as nn
import torch.nn.functional as F

NC = 3
IMG_SIZE = 128
DIM_EMBED = 128


#------------------------------------------------------------------------------
class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet_embed(nn.Module):
    def __init__(self, block, num_blocks, nc=NC, dim_embed=DIM_EMBED):
        super(ResNet_embed, self).__init__()
        self.in_planes = 64

        self.main = nn.Sequential(
            nn.Conv2d(nc, 64, kernel_size=3, stride=1, padding=1, bias=False),  # h=h
            # nn.Conv2d(nc, 64, kernel_size=4, stride=2, padding=1, bias=False),  # h=h/2
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2,2), #h=h/2 64
            # self._make_layer(block, 64, num_blocks[0], stride=1),  # h=h
            self._make_layer(block, 64, num_blocks[0], stride=2),  # h=h/2 32
            self._make_layer(block, 128, num_blocks[1], stride=2), # h=h/2 16
            self._make_layer(block, 256, num_blocks[2], stride=2), # h=h/2 8
            self._make_layer(block, 512, num_blocks[3], stride=2), # h=h/2 4
            # nn.AvgPool2d(kernel_size=4)
            nn.AdaptiveAvgPool2d((1, 1))
        )

        self.x2h_res = nn.Sequential(
            nn.Linear(512, 512),
            nn.BatchNorm1d(512),
            nn.ReLU(),

            nn.Linear(512, dim_embed),
            nn.BatchNorm1d(dim_embed),
            nn.ReLU(),
        )

        self.h2y = nn.Sequential(
            nn.Linear(dim_embed, 1),
            nn.ReLU()
        )

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):

        features = self.main(x)
        features = features.view(features.size(0), -1)
        features = self.x2h_res(features)
        out = self.h2y(features)

        return out, features


def ResNet18_embed(dim_embed=DIM_EMBED):
    return ResNet_embed(BasicBlock, [2,2,2,2], dim_embed=dim_embed)

def ResNet34_embed(dim_embed=DIM_EMBED):
    return ResNet_embed(BasicBlock, [3,4,6,3], dim_embed=dim_embed)

def ResNet50_embed(dim_embed=DIM_EMBED):
    return ResNet_embed(Bottleneck, [3,4,6,3], dim_embed=dim_embed)

#------------------------------------------------------------------------------
# map labels to the embedding space
class model_y2h(nn.Module):
    def __init__(self, dim_embed=DIM_EMBED):
        super(model_y2h, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(1, dim_embed),
            # nn.BatchNorm1d(dim_embed),
            nn.GroupNorm(8, dim_embed),
            nn.ReLU(),

            nn.Linear(dim_embed, dim_embed),
            # nn.BatchNorm1d(dim_embed),
            nn.GroupNorm(8, dim_embed),
            nn.ReLU(),

            nn.Linear(dim_embed, dim_embed),
            # nn.BatchNorm1d(dim_embed),
            nn.GroupNorm(8, dim_embed),
            nn.ReLU(),

            nn.Linear(dim_embed, dim_embed),
            # nn.BatchNorm1d(dim_embed),
            nn.GroupNorm(8, dim_embed),
            nn.ReLU(),

            nn.Linear(dim_embed, dim_embed),
            nn.ReLU()
        )

    def forward(self, y):
        y = y.view(-1, 1) +1e-8
        # y = torch.exp(y.view(-1, 1))
        return self.main(y)



if __name__ == "__main__":
    net = ResNet34_embed(dim_embed=128).cuda()
    x = torch.randn(16,NC,IMG_SIZE,IMG_SIZE).cuda()
    out, features = net(x)
    print(out.size())
    print(features.size())

    net_y2h = model_y2h()
